import json
import os.path
import tempfile
import unittest
from unittest import mock

from financeager import DEFAULT_POCKET_NAME, DEFAULT_TABLE, RECURRENT_TABLE, convert

TEST_DATA_DIR = tempfile.mkdtemp(prefix="financeager-")


@mock.patch("financeager.DATA_DIR", TEST_DATA_DIR)
class ConvertTestCase(unittest.TestCase):
    POCKET_FILEPATH = os.path.join(TEST_DATA_DIR, "{}.json".format(DEFAULT_POCKET_NAME))

    def tearDown(self):
        if os.path.exists(self.POCKET_FILEPATH):
            os.remove(self.POCKET_FILEPATH)

    def test_convert_no_period(self):
        convert.run(period_filepaths=[])
        self.assertTrue(os.path.exists(self.POCKET_FILEPATH))

        with open(self.POCKET_FILEPATH) as f:
            self.assertEqual(f.read(), "")

    def test_convert_empty_period(self):
        expected_content = {DEFAULT_TABLE: {}, RECURRENT_TABLE: {}}

        period_filepath = os.path.join(TEST_DATA_DIR, "1234.json")
        with open(period_filepath, "w") as f:
            json.dump(expected_content, f)
        convert.run(period_filepaths=[period_filepath])

        self.assertTrue(os.path.exists(self.POCKET_FILEPATH))

        with open(self.POCKET_FILEPATH) as f:
            actual_content = json.load(f)
        self.assertDictEqual(actual_content, expected_content)

    def test_convert_single_period(self):
        period_content = {
            DEFAULT_TABLE: {
                "1": {
                    "value": 42.0,
                    "category": None,
                    "name": "entry",
                    "date": "12-28",
                },
            },
            RECURRENT_TABLE: {
                "1": {
                    "value": -500.0,
                    "category": None,
                    "name": "rent",
                    "start": "01-01",
                    "end": "12-31",
                    "frequency": "monthly",
                },
            },
        }

        period_filepath = os.path.join(TEST_DATA_DIR, "1235.json")
        with open(period_filepath, "w") as f:
            json.dump(period_content, f)
        convert.run(period_filepaths=[period_filepath])

        self.assertTrue(os.path.exists(self.POCKET_FILEPATH))

        expected_content = period_content.copy()
        expected_content[DEFAULT_TABLE]["1"]["date"] = "1235-12-28"
        expected_content[RECURRENT_TABLE]["1"]["start"] = "1235-01-01"
        expected_content[RECURRENT_TABLE]["1"]["end"] = "1235-12-31"

        with open(self.POCKET_FILEPATH) as f:
            actual_content = json.load(f)

        self.assertDictEqual(actual_content, expected_content)

    def test_convert_multiple_periods(self):
        period_content = {
            DEFAULT_TABLE: {
                "1": {
                    "value": 42.0,
                    "category": None,
                    "name": "entry",
                    "date": "12-28",
                },
            },
            RECURRENT_TABLE: {
                "1": {
                    "value": -500.0,
                    "category": None,
                    "name": "rent",
                    "start": "01-01",
                    "end": "12-31",
                    "frequency": "monthly",
                },
            },
        }
        period_filepath = os.path.join(TEST_DATA_DIR, "1236.json")
        with open(period_filepath, "w") as f:
            json.dump(period_content, f)
        period_filepaths = [period_filepath]

        expected_content = period_content.copy()
        expected_content[DEFAULT_TABLE]["1"]["date"] = "1236-12-28"
        expected_content[RECURRENT_TABLE]["1"]["start"] = "1236-01-01"
        expected_content[RECURRENT_TABLE]["1"]["end"] = "1236-12-31"

        period_content = {
            DEFAULT_TABLE: {
                "1": {
                    "value": -5.0,
                    "category": "food",
                    "name": "burrito",
                    "date": "03-04",
                },
            },
        }
        period_filepath = os.path.join(TEST_DATA_DIR, "1237.json")
        with open(period_filepath, "w") as f:
            json.dump(period_content, f)
        period_filepaths.append(period_filepath)

        expected_content[DEFAULT_TABLE]["2"] = period_content[DEFAULT_TABLE]["1"].copy()
        expected_content[DEFAULT_TABLE]["2"]["date"] = "1237-03-04"

        convert.run(period_filepaths=period_filepaths)

        self.assertTrue(os.path.exists(self.POCKET_FILEPATH))

        with open(self.POCKET_FILEPATH) as f:
            actual_content = json.load(f)

        self.assertDictEqual(actual_content, expected_content)


if __name__ == "__main__":
    unittest.main()



"""Data structures for smallest elements of frontend representation of database
query results."""

import time

from . import DEFAULT_BASE_ENTRY_SORT_KEY, POCKET_DATE_FORMAT


class Entry:
    """Base class. An entry represents a row in the table that is built from a
    Model.
    The name field is stored in lowercase, simplifying searching from the parent
    listing. The value is rendered absolute to simplify sorting."""

    def __init__(self, name, value):
        """:type name: str
        :type value: float or int
        """
        self.name = name.lower()
        self.value = abs(value)


class BaseEntry(Entry):
    """Innermost element of the Model, child of a CategoryEntry. Holds
    information on name, value, date and eid."""

    ITEM_TYPES = ["name", "value", "date"]

    NAME_LENGTH = 16
    VALUE_LENGTH = 8  # 00000.00
    VALUE_DIGITS = 2
    DATE_FORMAT = "%y-%m-%d"
    DATE_LENGTH = 8  # yy-mm-dd
    SHOW_EID = True
    EID_LENGTH = 4 if SHOW_EID else 0
    # add spaces separating name/value, value/date and date/eid
    TOTAL_LENGTH = (
        NAME_LENGTH + VALUE_LENGTH + DATE_LENGTH + EID_LENGTH + 3 if SHOW_EID else 2
    )

    def __init__(self, name, value, date, eid=0):
        """:type eid: int or string, will be converted to int
        :type date: str of valid format
        """
        super().__init__(name, value)
        self.date = time.strftime(
            self.DATE_FORMAT, time.strptime(date, POCKET_DATE_FORMAT)
        )
        self.eid = int(eid)

    def __str__(self):
        """Return a formatted string representing the entry."""

        string = "{name:{0}.{0}} {value:>{1}.{2}f} {date}".format(
            self.NAME_LENGTH,
            self.VALUE_LENGTH,
            self.VALUE_DIGITS,
            name=self.name.title(),
            value=self.value,
            date=self.date,
        )
        if self.SHOW_EID:
            string += " {1:{0}d}".format(self.EID_LENGTH, self.eid)
        return string


class CategoryEntry(Entry):
    """First child of the listing, holding BaseEntries. Has a name and a value
    (i.e. the sum of its children's values)."""

    ITEM_TYPES = ["name", "sum", "empty"]
    DEFAULT_NAME = "unspecified"

    BASE_ENTRY_INDENT = 2
    NAME_LENGTH = BaseEntry.NAME_LENGTH + BASE_ENTRY_INDENT
    TOTAL_LENGTH = BaseEntry.TOTAL_LENGTH + BASE_ENTRY_INDENT

    def __init__(self, name, entries=None):
        """:type entries: list[BaseEntry]"""
        super().__init__(name=name or self.DEFAULT_NAME, value=0.0)

        self.entries = []
        if entries is not None:
            for base_entry in entries:
                self.append(base_entry)

    def append(self, base_entry):
        """Append a BaseEntry to the category and update the value."""
        self.entries.append(base_entry)
        self.value += base_entry.value

    def string(self, *, entry_sort=None, total_listing_value=None):
        """Return a formatted string representing the entry including its
        children (i.e. BaseEntries). The category representation is supposed
        to be longer than the BaseEntry representation so that the latter is
        indented.
        'entry_sort' determines the property according to which the list of base
        entries is sorted.
        If 'total_listing_value' is set, BaseEntries are not listed but instead
        the share in the total listing of the category is inserted.
        """
        percent = ""
        if total_listing_value is not None:
            percent = "{1:>{0}.1f}".format(
                BaseEntry.DATE_LENGTH, 100 * self.value / total_listing_value
            )

        lines = [
            "{name:{0}.{0}} {value:>{1}.{2}f} {percent}".format(
                self.NAME_LENGTH,
                BaseEntry.VALUE_LENGTH,
                BaseEntry.VALUE_DIGITS,
                name=self.name.title(),
                value=self.value,
                percent=percent,
            ).ljust(self.TOTAL_LENGTH)
        ]

        if total_listing_value is not None:
            return lines[0]

        entry_sort = entry_sort or DEFAULT_BASE_ENTRY_SORT_KEY
        sort_key = lambda e: getattr(e, entry_sort)
        for entry in sorted(self.entries, key=sort_key):
            lines.append(self.BASE_ENTRY_INDENT * " " + str(entry))

        return "\n".join(lines)


def prettify(element, *, default_category):
    """Return element properties formatted as list. The type of the element
    (recurrent or standard) is inferred by the presence of the 'frequency'
    property.
    If the element's 'category' property is None, use 'default_category'.

    :type element: dict
    :type default_category: str
    """
    recurrent = "frequency" in element

    # Define order of listed properties
    if recurrent:
        properties = ("name", "value", "frequency", "start", "end", "category")
    else:
        properties = ("name", "value", "date", "category")
    longest_property_length = 0
    for p in properties:
        if len(p) > longest_property_length:
            longest_property_length = len(p)

    if element["category"] is None:
        element["category"] = default_category

    lines = []
    for p in properties:
        lines.append(
            "{}: {}".format(
                p.capitalize().ljust(longest_property_length), str(element[p]).title()
            )
        )

    return "\n".join(lines)



"""Tabular, frontend-representation of financeager pocket."""
from . import DEFAULT_CATEGORY_ENTRY_SORT_KEY, DEFAULT_TABLE, RECURRENT_TABLE
from .entries import BaseEntry, CategoryEntry


class Listing:
    """Holds Entries in hierarchical order. First-level children are
    CategoryEntries, second-level children are BaseEntries. Generator methods
    are provided to iterate over these."""

    def __init__(self, name=None, categories=None):
        self.name = name or "Listing"
        self.categories = categories or []

    @classmethod
    def from_elements(cls, elements, default_category=None, name=None):
        """Create listing from list of element dictionaries"""
        listing = cls(name=name)
        for element in elements:
            category = element.pop("category", None) or default_category
            listing.add_entry(BaseEntry(**element), category_name=category)
        return listing

    def prettify(
        self, *, category_sort=None, category_percentage=False, **entry_options
    ):
        """Format listing (incl. name and header).
        Category entries are sorted acc. to the given 'category_sort'.
        'entry_options' are passed to CategoryEntry.string().
        The header lists the relevant item types of BaseEntry, or, if
        'category_percentage' is set, an indicator for a column that displays
        the share in the listing total of each category.

        :return: str
        """
        result = ["{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, self.name)]

        if category_percentage:
            entry_options["total_listing_value"] = self.total_value()

            # yapf: disable
            header_line = "{3:{0}} {4:{1}} {5:>{2}} {6}".format(
                CategoryEntry.NAME_LENGTH, BaseEntry.VALUE_LENGTH,
                BaseEntry.DATE_LENGTH,
                *[k.capitalize() for k in BaseEntry.ITEM_TYPES[:2]], "%",
                BaseEntry.EID_LENGTH * " ")
            # yapf: enable

        else:
            header_line = "{3:{0}} {4:{1}} {5:{2}}".format(
                CategoryEntry.NAME_LENGTH,
                BaseEntry.VALUE_LENGTH,
                BaseEntry.DATE_LENGTH,
                *[k.capitalize() for k in BaseEntry.ITEM_TYPES]
            )
            if BaseEntry.SHOW_EID:
                header_line += " " + "ID".ljust(BaseEntry.EID_LENGTH)

        result.append(header_line)

        category_sort = category_sort or DEFAULT_CATEGORY_ENTRY_SORT_KEY
        sort_key = lambda e: getattr(e, category_sort)
        for category in sorted(self.categories, key=sort_key):
            result.append(category.string(**entry_options))

        return "\n".join(result)

    def add_entry(self, entry, category_name=None):
        """Add a Category- or BaseEntry to the listing.
        Category names are unique, i.e. a CategoryEntry is discarded if one
        with identical name (case INsensitive) already exists.
        When adding a BaseEntry, the parent CategoryEntry is created if it does
        not exist. If no category is specified, the BaseEntry is added to the
        default category.

        :raises: TypeError if neither CategoryEntry nor BaseEntry given
        """
        if isinstance(entry, CategoryEntry):
            if entry.name not in self.category_entry_names:
                self.categories.append(entry)
        elif isinstance(entry, BaseEntry):
            category_entry = self._get_category_entry(category_name)
            category_entry.append(entry)
        else:
            raise TypeError("Invalid entry type: {}".format(entry))

    def category_fields(self, field_type):
        """Generator iterating over the field specified by `field_type` of the
        first-level children (CategoryEntries) of the listing.

        :param field_type: 'name' or 'value'

        raises: KeyError if `field_type` not found.
        yields: str or float
        """
        for category_entry in self.categories:
            yield getattr(category_entry, field_type)

    @property
    def category_entry_names(self):
        """Convenience generator method yielding category names."""
        for category_name in self.category_fields("name"):
            yield category_name

    def _get_category_entry(self, category_name):
        """Fetch CategoryEntry searching for given `category_name` or return a
        new instance that is automatically added to the Listing's categories.
        The search is case insensitive.

        :return: CategoryEntry
        """
        category_name = category_name.lower()

        for category_entry in self.categories:
            if category_entry.name == category_name:
                return category_entry
        else:
            # Nothing found in existing categories
            category_entry = CategoryEntry(name=category_name)
            self.add_entry(category_entry)
            return category_entry

    def total_value(self):
        """Return total value of the listing."""
        return sum(v for v in self.category_fields("value"))


def prettify(elements, stacked_layout=False, recurrent_only=False, **listing_options):
    """Sort the given elements (type acc. to Pocket._search_all_tables) by
    positive and negative value and return pretty string build from the
    corresponding Listings.

    :param stacked_layout: If True, listings are displayed one by one
    :param recurrent_only: If True, assume that given elements are purely
        recurrent ones, and return them formatted as table
    :param listing_options: Options passed to Listing.prettify(), and
        Listing.from_elements()
    """

    if recurrent_only:
        fields = ["id", "name", "value", "category", "start", "end", "frequency"]
        field_lengths = {f: len(f) for f in fields}
        # Determine max. field length and convert some field types
        for element in elements:
            element["category"] = element["category"] or "Unspecified"
            element["end"] = element["end"] or "-"
            for field, length in field_lengths.items():
                field_lengths[field] = max(length, len(str(element[field])))

        sep = " | "
        # Add all-uppercase header row
        lines = [sep.join(f.upper().ljust(l) for f, l in field_lengths.items())]

        def _format(element, field):
            # Numeric values shall be right-aligned, others left-aligned
            if field in ["id", "value"]:
                return str(element[field]).rjust(field_lengths[field])
            return element[field].capitalize().ljust(field_lengths[field])

        # Sort elements acc. to 'entry_sort' option, and format them
        entry_sort = listing_options.get("entry_sort") or "id"
        entry_sort = "id" if entry_sort == "eid" else entry_sort
        for element in sorted(elements, key=lambda e: e[entry_sort]):
            lines.append(sep.join(_format(element, f) for f in fields))

        return "\n".join(lines)

    earnings = []
    expenses = []

    def _sort(eid, element):
        # Copying avoids modifying the original element. Flattening is in order
        # to distinguish recurrent entries (they have the same element ID which
        # thus can't be used as dict key)
        flat_element = element.copy()
        flat_element["eid"] = eid
        if flat_element["value"] > 0:
            earnings.append(flat_element)
        else:
            expenses.append(flat_element)

    # process standard elements
    for eid, element in elements[DEFAULT_TABLE].items():
        _sort(eid, element)

    # process recurrent elements, i.e. for each eid iterate list
    for eid, recurrent_elements in elements[RECURRENT_TABLE].items():
        for element in recurrent_elements:
            _sort(eid, element)

    if not earnings and not expenses:
        return ""

    default_category = listing_options.pop("default_category", None)
    listing_earnings = Listing.from_elements(
        earnings, default_category=default_category, name="Earnings"
    )
    listing_expenses = Listing.from_elements(
        expenses, default_category=default_category, name="Expenses"
    )
    listings = [listing_earnings, listing_expenses]

    total_values = []
    total_entries = []
    for listing in listings:
        total_entry = CategoryEntry(name="TOTAL")
        total_entry.value = listing.total_value()
        total_values.append(total_entry.value)
        total_entries.append(total_entry.string())

    # Compute difference of total earnings and expenses
    diff_entry = CategoryEntry(name="Difference")
    diff_entry.value = total_values[0] - total_values[1]
    diff_entry = diff_entry.string()

    if stacked_layout:
        line = CategoryEntry.TOTAL_LENGTH * "="
        return "{}\n{}\n{}\n\n{}\n{}\n{}\n{}".format(
            listing_earnings.prettify(**listing_options),
            line,
            total_entries[0],
            listing_expenses.prettify(**listing_options),
            line,
            total_entries[1],
            diff_entry,
        )
    else:
        result = []
        listings_str = [ls.prettify(**listing_options).splitlines() for ls in listings]
        for row in zip(*listings_str):
            result.append(" | ".join(row))
        earnings_size = len(listings_str[0])
        expenses_size = len(listings_str[1])
        diff = earnings_size - expenses_size
        if diff > 0:
            for row in listings_str[0][expenses_size:]:
                result.append(row + " | ")
        else:
            for row in listings_str[1][earnings_size:]:
                result.append(CategoryEntry.TOTAL_LENGTH * " " + " | " + row)
        # add 3 to take central separator " | " into account
        result.append((2 * CategoryEntry.TOTAL_LENGTH + 3) * "=")

        # add total value of earnings and expenses, and difference thereof
        result.append(" | ".join(total_entries))
        result.append(diff_entry)

        return "\n".join(result)



import unittest

from financeager import DEFAULT_TABLE, RECURRENT_TABLE
from financeager.entries import BaseEntry, CategoryEntry
from financeager.listing import Listing, prettify


class AddCategoryEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.category_name = "Groceries"
        self.listing.add_entry(CategoryEntry(name=self.category_name))

    def test_category_item_in_list(self):
        self.assertIn(self.category_name.lower(), self.listing.category_entry_names)


class AddCategoryEntryTwiceTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.category_name = "Groceries"
        self.listing.add_entry(CategoryEntry(name=self.category_name))
        self.listing.add_entry(CategoryEntry(name=self.category_name))

    def test_category_item_in_list(self):
        self.assertIn(self.category_name.lower(), self.listing.category_entry_names)

    def test_single_item_in_list(self):
        self.assertEqual(1, len(list(self.listing.category_entry_names)))


class AddBaseEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_name = "Aldi"
        self.item_value = 66.6
        self.item_date = "2000-11-08"
        self.item_category = "Groceries"
        self.listing.add_entry(
            BaseEntry(self.item_name, self.item_value, self.item_date),
            self.item_category,
        )

    def test_str(self):
        self.assertEqual(
            self.listing.prettify(),
            "\n".join(
                [
                    "{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, "Listing"),
                    "Name               Value    Date     ID  ",
                    "Groceries             66.60" + 14 * " ",
                    "  Aldi                66.60 00-11-08    0",
                ]
            ),
        )

    def test_str_no_eid(self):
        BaseEntry.SHOW_EID = False
        self.assertEqual(
            self.listing.prettify(),
            "\n".join(
                [
                    "{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, "Listing"),
                    "Name               Value    Date    ",
                    # TODO: fix this; category entry line has to be shorter
                    "Groceries             66.60              ",
                    "  Aldi                66.60 00-11-08",
                ]
            ),
        )
        BaseEntry.SHOW_EID = True

    def test_add_invalid_entry(self):
        self.assertRaises(TypeError, self.listing.add_entry, None)


class SortCategoryEntriesTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        for c, v in zip("ab", [20, 10]):
            self.listing.add_entry(BaseEntry("foo", v, "2000-01-01"), c)

    def test_sort_by_name(self):
        self.assertEqual(
            self.listing.prettify(category_sort="name"),
            "\n".join(
                [
                    "{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, "Listing"),
                    "Name               Value    Date     ID  ",
                    "A                     20.00" + 14 * " ",
                    "  Foo                 20.00 00-01-01    0",
                    "B                     10.00" + 14 * " ",
                    "  Foo                 10.00 00-01-01    0",
                ]
            ),
        )

    def test_sort_by_value(self):
        self.assertEqual(
            self.listing.prettify(),
            "\n".join(
                [
                    "{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, "Listing"),
                    "Name               Value    Date     ID  ",
                    "B                     10.00" + 14 * " ",
                    "  Foo                 10.00 00-01-01    0",
                    "A                     20.00" + 14 * " ",
                    "  Foo                 20.00 00-01-01    0",
                ]
            ),
        )


class AddNegativeBaseEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_name = "Aldi"
        self.item_value = -66.6
        self.item_date = "2000-11-08"
        self.item_category = "Groceries"
        self.listing.add_entry(
            BaseEntry(self.item_name, self.item_value, self.item_date),
            self.item_category,
        )

    def test_str(self):
        self.assertEqual(
            self.listing.prettify(),
            "\n".join(
                [
                    "{1:^{0}}".format(CategoryEntry.TOTAL_LENGTH, "Listing"),
                    "Name               Value    Date     ID  ",
                    "Groceries             66.60" + 14 * " ",
                    "  Aldi                66.60 00-11-08    0",
                ]
            ),
        )


class AddBaseEntryWithoutCategoryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_name = "Aldi"
        self.item_value = 66.6
        self.item_date = "2009-11-08"
        self.listing.add_entry(
            BaseEntry(self.item_name, self.item_value, self.item_date),
            category_name=CategoryEntry.DEFAULT_NAME,
        )

    def test_default_category_in_list(self):
        names = list(self.listing.category_entry_names)
        self.assertIn(CategoryEntry.DEFAULT_NAME, names)


class AddTwoBaseEntriesTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_a_value = 66.6
        self.item_b_value = 10.01
        self.item_category = "Groceries"
        self.date = "2008-11-11"
        self.listing.add_entry(
            BaseEntry("Aldi", self.item_a_value, self.date), self.item_category
        )
        self.listing.add_entry(
            BaseEntry("Rewe", self.item_b_value, self.date), self.item_category
        )

    def test_total_value(self):
        self.assertAlmostEqual(
            self.item_a_value + self.item_b_value, self.listing.total_value(), places=5
        )


class ListingFromElementsTestCase(unittest.TestCase):
    def setUp(self):
        self.name = "Dinner for one"
        self.value = 99.9
        self.date = "1980-12-31"
        self.listing = Listing.from_elements(
            [dict(name=self.name, value=self.value, date=self.date, eid=0)],
            default_category=CategoryEntry.DEFAULT_NAME,
        )

    def test_contains_an_entry(self):
        self.assertIn(self.date[2:], self.listing.prettify())

    def test_category_item_names(self):
        parsed_listing_entry_names = list(self.listing.category_entry_names)
        listing_entry_names = [CategoryEntry.DEFAULT_NAME]
        self.assertListEqual(listing_entry_names, parsed_listing_entry_names)


class PrettifyListingsTestCase(unittest.TestCase):
    def test_prettify_no_elements(self):
        elements = {DEFAULT_TABLE: {}, RECURRENT_TABLE: {}}
        self.assertEqual(prettify(elements), "")

    def test_prettify(self):
        elements = {
            DEFAULT_TABLE: {
                1: {
                    "name": "food",
                    "value": -100.01,
                    "date": "2000-03-03",
                    "category": "groceries",
                },
                999: {"name": "money", "value": 299.99, "date": "2000-03-03"},
            },
            RECURRENT_TABLE: {
                42: [
                    {
                        "name": "gold",
                        "value": 4321,
                        "date": "2000-01-01",
                        "category": "bank",
                    }
                ]
            },
        }
        self.maxDiff = None
        elements_copy = elements.copy()
        self.assertEqual(
            prettify(elements_copy, default_category=CategoryEntry.DEFAULT_NAME),
            "                Earnings                  |                 Expenses                 \n"  # noqa
            "Name               Value    Date     ID   | Name               Value    Date     ID  \n"  # noqa
            "Unspecified          299.99               | Groceries            100.01              \n"  # noqa
            "  Money              299.99 00-03-03  999 |   Food               100.01 00-03-03    1\n"  # noqa
            "Bank                4321.00               | \n"  # noqa
            "  Gold              4321.00 00-01-01   42 | \n"  # noqa
            "=====================================================================================\n"  # noqa
            "Total               4620.99               | Total                100.01              \n"  # noqa
            "Difference          4520.98              ",
        )
        # Assert that original data was not modified
        self.assertDictEqual(elements, elements_copy)

        self.assertEqual(
            prettify(
                elements_copy,
                default_category=CategoryEntry.DEFAULT_NAME,
                category_percentage=True,
            ),
            "                Earnings                  |                 Expenses                 \n"  # noqa
            "Name               Value           %      | Name               Value           %     \n"  # noqa
            "Unspecified          299.99      6.5      | Groceries            100.01    100.0     \n"  # noqa
            "Bank                4321.00     93.5      | \n"  # noqa
            "=====================================================================================\n"  # noqa
            "Total               4620.99               | Total                100.01              \n"  # noqa
            "Difference          4520.98              ",
        )

    def test_prettify_stacked_layout(self):
        elements = {
            DEFAULT_TABLE: {
                2: {
                    "name": "shirt",
                    "value": -199,
                    "date": "2000-04-01",
                    "category": "clothes",
                },
                3: {
                    "name": "lunch",
                    "value": -20,
                    "date": "2000-04-01",
                    "category": "food",
                },
            },
            RECURRENT_TABLE: {},
        }
        self.maxDiff = None
        self.assertEqual(
            prettify(elements, stacked_layout=True),
            "\
                Earnings                 "
            + "\n\
Name               Value    Date     ID  "
            + "\n\
========================================="
            + "\n\
Total                  0.00              "
            + """

"""
            + "\
                Expenses                 "
            + "\n\
Name               Value    Date     ID  "
            + "\n\
Food                  20.00              "
            + "\n\
  Lunch               20.00 00-04-01    3"
            + "\n\
Clothes              199.00              "
            + "\n\
  Shirt              199.00 00-04-01    2"
            + "\n\
========================================="
            + "\n\
Total                219.00              "
            + "\n\
Difference          -219.00              ",
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from financeager.entries import BaseEntry, CategoryEntry
from financeager.entries import prettify as prettify_entry


class BaseEntryTestCase(unittest.TestCase):
    def test_attributes(self):
        entry = BaseEntry(
            **{"name": "groceries", "value": 123.45, "date": "2000-08-10"}
        )

        self.assertEqual(entry.name, "groceries")
        self.assertAlmostEqual(entry.value, 123.45, places=5)
        self.assertEqual(entry.date, "00-08-10")
        self.assertEqual(entry.eid, 0)

    def test_leap_year_date(self):
        entry = BaseEntry(
            **{
                "name": "leap",
                "value": 1,
                "date": "2000-02-29",
            }
        )
        self.assertEqual(entry.date, "00-02-29")


class NegativeBaseEntryTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.entry = BaseEntry(
            **{"name": "vw bully", "value": -6000, "date": "2000-01-01"}
        )

    def test_name(self):
        self.assertEqual(self.entry.name, "vw bully")

    def test_value(self):
        self.assertEqual(self.entry.value, 6000)

    def test_str(self):
        expected = "Vw Bully".ljust(BaseEntry.NAME_LENGTH) + " " + " 6000.00 00-01-01"
        expected += "    0"
        self.assertEqual(str(self.entry), expected)

    def test_str_no_eid(self):
        expected = "Vw Bully".ljust(BaseEntry.NAME_LENGTH) + " " + " 6000.00 00-01-01"
        BaseEntry.SHOW_EID = False
        self.assertEqual(str(self.entry), expected)
        BaseEntry.SHOW_EID = True


class CategoryEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.entry = CategoryEntry(name="gifts")

    def test_name(self):
        self.assertEqual(self.entry.name, "gifts")

    def test_value(self):
        self.assertEqual(self.entry.value, 0.0)

    def test_str(self):
        self.assertEqual(
            self.entry.string(),
            "Gifts".ljust(CategoryEntry.NAME_LENGTH)
            + " "
            + "0.00".rjust(BaseEntry.VALUE_LENGTH)
            + " "
            + BaseEntry.DATE_LENGTH * " "
            + (BaseEntry.EID_LENGTH + 1) * " "
            if BaseEntry.SHOW_EID
            else "",
        )


class LongNegativeCategoryEntryTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.entry = CategoryEntry(
            name="This is quite a LOOONG Category",
            entries=[BaseEntry("entry", -100, "2000-08-13")],
        )

    def test_name(self):
        self.assertEqual(self.entry.name, "this is quite a looong category")

    def test_value(self):
        self.assertEqual(self.entry.value, 100)

    def test_str(self):
        self.assertEqual(
            self.entry.string(),
            "This Is Quite A Lo "
            + "  100.00"
            + 9 * " "
            + 5 * " "
            + "\n"
            + "  Entry            "
            + "  100.00"
            + " 00-08-13 "
            + "   0",
        )
        self.assertEqual(
            self.entry.string(total_listing_value=200),
            "This Is Quite A Lo " + "  100.00" + "     50.0" + 5 * " ",
        )


class SortBaseEntriesTestCase(unittest.TestCase):
    def setUp(self):
        self.entry = CategoryEntry(
            name="letters",
            entries=[
                BaseEntry(lt, v, "2000-0{}-11".format(9 - v), i)
                for lt, v, i in zip("abc", (1, 5, 3), (7, 1, 3))
            ],
        )

    def test_sort_by_name(self):
        # don't let trailing whitespaces of category name row being cleaned up
        self.assertEqual(
            self.entry.string(entry_sort="name"),
            "\
Letters                9.00              "
            + """
  A                    1.00 00-08-11    7
  B                    5.00 00-04-11    1
  C                    3.00 00-06-11    3""",
        )

    def test_sort_by_value(self):
        self.assertEqual(
            self.entry.string(entry_sort="value"),
            "\
Letters                9.00              "
            + """
  A                    1.00 00-08-11    7
  C                    3.00 00-06-11    3
  B                    5.00 00-04-11    1""",
        )

    def test_sort_by_date(self):
        self.assertEqual(
            self.entry.string(entry_sort="date"),
            "\
Letters                9.00              "
            + """
  B                    5.00 00-04-11    1
  C                    3.00 00-06-11    3
  A                    1.00 00-08-11    7""",
        )

    def test_sort_by_eid(self):
        self.assertEqual(
            self.entry.string(entry_sort="eid"),
            "\
Letters                9.00              "
            + """
  B                    5.00 00-04-11    1
  C                    3.00 00-06-11    3
  A                    1.00 00-08-11    7""",
        )


class PrettifyBaseEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.element = dict(
            name="soccer shoes", value=-123.45, date="04-01", category="sport equipment"
        )

    def test_prettify(self):
        self.assertEqual(
            prettify_entry(self.element, default_category=CategoryEntry.DEFAULT_NAME),
            """\
Name    : Soccer Shoes
Value   : -123.45
Date    : 04-01
Category: Sport Equipment""",
        )

    def test_prettify_default_category(self):
        element = self.element.copy()
        element["category"] = None
        self.assertEqual(
            prettify_entry(element, default_category=CategoryEntry.DEFAULT_NAME),
            """\
Name    : Soccer Shoes
Value   : -123.45
Date    : 04-01
Category: {}""".format(
                CategoryEntry.DEFAULT_NAME.capitalize()
            ),
        )


class PrettifyRecurrentEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.element = dict(
            name="retirement money",
            value=567.0,
            category="income",
            frequency="monthly",
            start="01-01",
            end="12-31",
        )

    def test_prettify(self):
        self.assertEqual(
            prettify_entry(self.element, default_category=CategoryEntry.DEFAULT_NAME),
            """\
Name     : Retirement Money
Value    : 567.0
Frequency: Monthly
Start    : 01-01
End      : 12-31
Category : Income""",
        )


if __name__ == "__main__":
    unittest.main()



"""Conversion of 'period' into 'pocket' database format."""
import glob
import os.path

from tinydb import database

import financeager

from . import exceptions, pocket


def main(*, sinks, period_filepaths=None):
    """Validate given period filepaths and run conversion. If no filepaths
    given, use all JSON files in `financeager.DATA_DIR`.
    Return True on success, otherwise False.
    """
    if period_filepaths is None:
        period_filepaths = sorted(
            glob.glob(os.path.join(financeager.DATA_DIR, "*.json"))
        )
    else:
        non_existing_filepaths = [f for f in period_filepaths if not os.path.exists(f)]
        if non_existing_filepaths:
            sinks.error(
                exceptions.ConversionError(
                    "One or more non-existing filepaths:\n{}".format(
                        "\n".join(non_existing_filepaths)
                    )
                )
            )
            return False

    period_filepaths = [
        f
        for f in period_filepaths
        if not f.endswith("{}.json".format(financeager.DEFAULT_POCKET_NAME))
    ]

    invalid_filepaths = []
    for filepath in period_filepaths:
        try:
            _extract_year_from_filepath(filepath)
        except ValueError:
            invalid_filepaths.append(filepath)
    if invalid_filepaths:
        sinks.error(
            exceptions.ConversionError(
                "One or more invalid filepaths:\n{}".format(
                    "\n".join(invalid_filepaths)
                )
            )
        )
        return False

    sinks.info("Converting {} period(s)...".format(len(period_filepaths)))
    run(period_filepaths)
    return True


def run(period_filepaths):
    """Read in period databases at given filepaths and create a single main
    Pocket from their contents. Date-related fields are updated with the
    according year.
    It has to be guaranteed that the period filepaths exist.
    """
    main_pocket = pocket.TinyDbPocket(data_dir=financeager.DATA_DIR)

    for filepath in period_filepaths:
        year = _extract_year_from_filepath(filepath)

        with database.TinyDB(filepath) as db:
            table_name = financeager.DEFAULT_TABLE
            documents = db.table(table_name).all()
            for doc in documents:
                doc["date"] = "{}-{}".format(year, doc["date"])
            main_pocket._db.table(table_name).insert_multiple(documents)

            table_name = financeager.RECURRENT_TABLE
            documents = db.table(table_name).all()
            for doc in documents:
                doc["start"] = "{}-{}".format(year, doc["start"])
                doc["end"] = "{}-{}".format(year, doc["end"])
            main_pocket._db.table(table_name).insert_multiple(documents)

    main_pocket.close()


def _extract_year_from_filepath(filepath):
    return int(os.path.splitext(os.path.basename(filepath))[0])
